@@ -1,4 +1,5 @@
 require "rexml/parent"
+require "rexml/parseexception"
 
 module REXML
 	##
@@ -7,13 +8,13 @@ module REXML
 	# being used to declare entities used in the document.
 	class DocType < Parent
 		START = "<!DOCTYPE"
-		START_RE = /\A\s*#{START}\s/m
+		START_RE = /\A\s*#{START}\s/um
 		STOP = ">"
-		STOP_RE = />/
+		STOP_RE = />/u
 		SYSTEM = "SYSTEM"
 		PUBLIC = "PUBLIC"
-		OPEN_RE = /\A\s*\[/
-		PATTERN_RE = /\s*#{START}\s+(.*?)(\[|>)/m
+		OPEN_RE = /\A\s*\[/u
+		PATTERN_RE = /\s*#{START}\s+(.*?)(\[|>)/um
 
 		## name is the name of the doctype
 		# external_id is the referenced DTD, if given
@@ -38,10 +39,10 @@ module REXML
 				identity = md[1]
 				close = md[2]
 
-				identity =~ /^(\w+)(\s+\w+)?(\s+["'].*?['"])?(\s+['"].*?["'])?/
+				identity =~ /^([!\*\w]+)(\s+\w+)?(\s+["'].*?['"])?(\s+['"].*?["'])?/u
 				@name = $1
 
-				raise "DOCTYPE is missing a name" if @name.nil?
+				raise ParseException.new("DOCTYPE is missing a name", first) if @name.nil?
 
 				@pub_sys = $2.nil? ? nil : $2.strip
 				@long_name = $3.nil? ? nil : $3.strip
@@ -101,7 +102,7 @@ module REXML
 			identity = md[1]
 			close = md[2]
 
-			identity =~ /^(\w+)(\s+\w+)?(\s+["'].*?['"])?(\s+['"].*?["'])?/
+			identity =~ /^(\w+)(\s+\w+)?(\s+["'].*?['"])?(\s+['"].*?["'])?/u
 			name = $1
 
 			raise "DOCTYPE is missing a name" if name.nil?
@@ -117,29 +118,42 @@ module REXML
 
 		private
 		def DocType.parser source
-			md = source.match(/\s*(.*?)>/m)
-			nxt = md[1]
-			until md[1] == "]" 
-				case md[1]
-				when /^%/ #/
-					md = source.match(/^\s*%(.*?);/m, true)
-					yield md[1]
-				when AttlistDecl::START_RE
-					yield AttlistDecl
-				when ElementDecl::START_RE
-					yield ElementDecl
-				when EntityDecl::START_RE
-					yield EntityDecl
-				when NotationDecl::START_RE
-					yield NotationDecl
-				when Comment::START_RE
-					yield Comment
-				when Instruction::START_RE
-					yield Instruction
-				else
-					raise "illegal entry #{md[1]} in DOCTYPE"
+			begin
+				md = source.match(/\s*(.*?)>/um)
+				nxt = md[1]
+				until md[1] == "]" 
+					case md[1]
+					when /^%/ #/u
+						md = source.match(/^\s*%(.*?);/um, true)
+						yield md[1]
+					when AttlistDecl::START_RE
+						yield AttlistDecl
+					when ElementDecl::START_RE
+						yield ElementDecl
+					when EntityDecl::START_RE
+						yield EntityDecl
+					when NotationDecl::START_RE
+						yield NotationDecl
+					when Comment::START_RE
+						yield Comment
+					when Instruction::START_RE
+						yield Instruction
+					else
+						if md.nil?
+							raise "no match!"
+						else
+							raise "illegal entry \"#{md[1]}\" in DOCTYPE"
+						end
+					end
+					md = source.match(/\s*(.*?)>/um)
+					raise ParseException.new( "Invalid end of DOCTYPE declaration \"#{source.buffer}\"", source ) if md.nil?
 				end
-				md = source.match(/\s*(.*?)>/m)
+				source.match(/\s*]>/um, true)
+			rescue ParseException
+				raise
+			rescue Exception => err
+				raise
+				raise ParseException.new( "Error parsing DOCTYPE declaration", source, nil, err )
 			end
 		end
 
@@ -195,8 +209,8 @@ module REXML
 	
 	class AttlistDecl < Declaration
 		START = "<!ATTLIST"
-		START_RE = /\s*#{START}/m
-		PATTERN_RE = /\s*(#{START}.*?>)/m
+		START_RE = /^\s*#{START}/um
+		PATTERN_RE = /\s*(#{START}.*?>)/um
 		def pattern
 			PATTERN_RE
 		end
@@ -204,28 +218,98 @@ module REXML
 
 	class ElementDecl < Declaration
 		START = "<!ELEMENT"
-		START_RE = /\s*#{START}/m
-		PATTERN_RE = /^s*(#{START}.*?>)/m
+		START_RE = /^\s*#{START}/um
+		PATTERN_RE = /^\s*(#{START}.*?)>/um
 		def pattern
 			PATTERN_RE
 		end
 	end
 
-	class EntityDecl < Declaration
+	class EntityDecl < Child
 		START = "<!ENTITY"
-		START_RE = /^\s*#{START}/m
-		PATTERN_RE = /^\s*(#{START}.*?>)/m
-		def pattern
-			PATTERN_RE
+		START_RE = /^\s*#{START}/um
+		PUBLIC = /^\s*#{START}\s+(?:%\s+)?(\w+)\s+PUBLIC\s+((["']).*?\3)\s+((["']).*?\5)\s*>/um
+		SYSTEM = /^\s*#{START}\s+(?:%\s+)?(\w+)\s+SYSTEM\s+((["']).*?\3)(?:\s+NDATA\s+\w+)?\s*>/um
+		PLAIN = /^\s*#{START}\s+(\w+)\s+((["']).*?\3)\s*>/um
+		PERCENT = /^\s*#{START}\s+%\s+(\w+)\s+((["']).*?\3)\s*>/um
+		# <!ENTITY name SYSTEM "...">
+		# <!ENTITY name "...">
+		def initialize src
+			super()
+			md = nil
+			if src.match( PUBLIC )
+				md = src.match( PUBLIC, true )
+				@middle = "PUBLIC"
+				@content = "#{md[2]} #{md[4]}"
+			elsif src.match( SYSTEM )
+				md = src.match( SYSTEM, true )
+				@middle = "SYSTEM"
+				@content = md[2]
+			elsif src.match( PLAIN )
+				md = src.match( PLAIN, true )
+				@middle = ""
+				@content = md[2]
+			elsif src.match( PERCENT )
+				md = src.match( PERCENT, true )
+				@middle = ""
+				@content = md[2]
+			end
+			raise ParseException.new("failed Entity match", src) if md.nil?
+			@name = md[1]
+		end
+
+		def to_s
+			rv = "<!ENTITY #@name "
+			rv << "#@middle " if @middle.size > 0
+			rv << @content
+			rv
+		end
+
+		def write( output, indent )
+			output << ('   '*indent) if indent > 0
+			output << to_s
+		end
+
+		def EntityDecl.parse_source source, listener
+			md = source.match( PATTERN_RE, true )
+			thing = md[0].squeeze " \t\n\r"
+			listener.send inspect.downcase, thing 
 		end
 	end
 
-	class NotationDecl < Declaration
+	class NotationDecl < Child
 		START = "<!NOTATION"
-		START_RE = /^\s*#{START}/m
-		PATTERN_RE = /^\s*(#{START}.*?>)/m
-		def pattern
-			PATTERN_RE
+		START_RE = /^\s*#{START}/um
+		#PATTERN_RE = /^\s*(#{START}.*?>)/um
+		PUBLIC = /^\s*#{START}\s+(\w[\w-]*)\s+(PUBLIC)\s+((["']).*?\4)\s*>/um
+		SYSTEM = /^\s*#{START}\s+(\w[\w-]*)\s+(SYSTEM)\s+((["']).*?\4)\s*>/um
+		def initialize src
+			super()
+			if src.match( PUBLIC )
+				md = src.match( PUBLIC, true )
+			elsif src.match( SYSTEM )
+				md = src.match( SYSTEM, true )
+			else
+				raise ParseException.new( "error parsing notation: no matching pattern", src )
+			end
+			@name = md[1]
+			@middle = md[2]
+			@rest = md[3]
+		end
+
+		def to_s
+			"<!NOTATION #@name #@middle #@rest>"
+		end
+
+		def write( output, indent )
+			output << ('   '*indent) if indent > 0
+			output << to_s
+		end
+
+		def NotationDecl.parse_source source, listener
+			md = source.match( PATTERN_RE, true )
+			thing = md[0].squeeze " \t\n\r"
+			listener.send inspect.downcase, thing 
 		end
 	end
 end
