@@ -3,6 +3,7 @@ require "rexml/namespace"
 require "rexml/attribute"
 require "rexml/cdata"
 require "rexml/xpath"
+require "rexml/parseexception"
 
 module REXML
 	# Represents a tagged XML element.  Elements are characterized by
@@ -16,6 +17,7 @@ module REXML
 		# Mechanisms for accessing attributes and child elements of this
 		# element.
 		attr_reader :attributes, :elements
+		attr_accessor :context
 
 		# Constructor
 		# @param parent if supplied, must be a Parent, and will be used as
@@ -26,11 +28,12 @@ module REXML
 		# and namespaces will be copied.
 		# If a Source, the source will be scanned and parsed for an Element,
 		# and all child elements will be recursively parsed as well.
-		def initialize( arg = UNDEFINED, parent=nil )
+		def initialize( arg = UNDEFINED, parent=nil, context=nil )
 			super(parent)
 
 			@elements = Elements.new self
 			@attributes = Attributes.new self
+			@context = context
 
 			if arg.kind_of? Source
 				parse arg
@@ -38,9 +41,10 @@ module REXML
 				self.name = arg
 			elsif arg.kind_of? Element
 				self.name = arg.expanded_name
-				arg.attributes.each{ |attribute|
+				arg.attributes.each_attribute{ |attribute|
 					@attributes << Attribute.new( attribute )
 				}
+				@context = arg.context
 			end
 		end
 
@@ -49,39 +53,55 @@ module REXML
 		end
 
 		# @return the root of the document that this element belongs to.
-		# If this element doesn't belong to a document, but does belong to
-		# another Element, the parent's root will be returned, until the
-		# earliest ancestor is found.
-		def root
-			parent.nil? ? self : parent.root
-		end
+	 # If this element doesn't belong to a document, but does belong to
+	 # another Element, the parent's root will be returned, until the
+	 # earliest ancestor is found.
+	 def root
+		 parent.nil? ? self : parent.root
+	 end
 
-		#################################################
-		# Namespaces                                    #
-		#################################################
+	 def whitespace
+		 @whitespace = (@context and @context[:respect_whitespace] and
+		 (@context[:respect_whitespace] == :all or
+		 @context[:respect_whitespace].include? expanded_name))
+		 @whitespace
+	 end
 
-		# @return an Array of all defined namespaces
-		def prefixes
-			prefixes = []
-			prefixes = parent.prefixes if parent
-			prefixes += attributes.prefixes
-			return prefixes
-		end
+	 def raw
+		 @raw = (@context and @context[:raw] and
+		 (@context[:raw] == :all or
+		 @context[:raw].include? expanded_name))
+		 @raw
+	 end
 
-		# @return the URI for a namespace, or the empty string if no such 
-		# namespace is declared
-		def namespace prefix=nil
-			if prefix.nil?
-				prefix = prefix()
-			end
-			prefix = "xmlns" if prefix == ""
-			ns = attributes[ prefix ]
-			ns = parent.namespace(prefix) if ns.nil? and parent
-			return ns
-		end
+	 once :whitespace, :raw
 
-		# Adds a namespace to this element
-		# @param prefix the namespace.  If the second argument is not supplied,
+	 #################################################
+	 # Namespaces                                    #
+	 #################################################
+
+	 # @return an Array of all defined namespaces
+	 def prefixes
+		 prefixes = []
+		 prefixes = parent.prefixes if parent
+		 prefixes += attributes.prefixes
+		 return prefixes
+	 end
+
+	 # @return the URI for a namespace, or the empty string if no such 
+	 # namespace is declared
+	 def namespace prefix=nil
+		 if prefix.nil?
+			 prefix = prefix()
+		 end
+		 prefix = "xmlns" if prefix == ""
+		 ns = attributes[ prefix ]
+		 ns = parent.namespace(prefix) if ns.nil? and parent
+		 return ns
+	 end
+
+	 # Adds a namespace to this element
+	 # @param prefix the namespace.  If the second argument is not supplied,
 		# this should be the URI of the default namespace.
 		# @param value (optional) the URI of the namespace.  If "", then unsets
 		# the default namespace
@@ -192,7 +212,7 @@ module REXML
 			element = element.previous_sibling until element.nil? or element.kind_of? Element
 			return element
 		end
-			
+
 
 		#################################################
 		# Text                                          #
@@ -245,7 +265,7 @@ module REXML
 		# @return self.  NOTE that contrary to most REXML methods, the replaced
 		# content is not returned.
 		def text=( text )
-			text = Text.new( text ) if text.kind_of? String
+			text = Text.new( text, whitespace(), raw() ) if text.kind_of? String
 			old_text = get_text
 			if text.nil?
 				old_text.remove unless old_text.nil?
@@ -317,7 +337,7 @@ module REXML
 		def write(writer, indent=0)
 			indent writer, indent
 			writer << "<#@expanded_name"
-			
+
 			@attributes.each_attribute do |attr|
 				writer << " "
 				attr.write( writer, indent )
@@ -328,28 +348,30 @@ module REXML
 			else
 				writer << ">"
 				write_children writer, indent
-				writer << "</#@name"
+				writer << "</#{expanded_name}"
 			end
 			writer << ">"
 		end
 
 		def Element.parse_stream source, listener
 			# Get the next tag
-			md = source.match(/^<([\w-]*:?[\w-]+)(.*?>)/m, true)
+			md = source.match(/^<([\w-]*:?[\w-]+)(.*?>)/um, true)
 			name = md[1]
 			rest = md[2]
 			attrs = rest.scan(Attribute::PATTERN)
+			attrs.each { |ar| ar[1,1] = [] }
 			listener.tag_start name, attrs
-			unless rest =~ /\/>$/		# empty
+			unless rest =~ /\/>$/u		# empty
 				while true
-					md = source.match(/\A(\s*[^>]*>)/m)
-					parse_error source if md.nil?
+					md = source.match(/\A(\s*[^>]*>)/um)
+					raise ParseException.new("malformed XML: bad content", source, 
+						self) unless md
 					line = md[1]
 					return if line.nil?
 					case line
-					when /\A<\w/m
+					when /\A<\w/um
 						Element.parse_stream source, listener
-					when /\A<\//m
+					when /\A<\//um
 						break
 					when CData::START_RE
 						CData.parse_stream source, listener
@@ -361,10 +383,11 @@ module REXML
 						Text.parse_stream source, listener
 					end
 				end
-				if source.match(/^\s*<\/#{name}\s*>/m, true)
+				if source.match(/^\s*<\/#{name}\s*>/um, true)
 					listener.tag_end name
 				else
-					parse_error source, "Missing end tag for #{expanded_name}"
+					raise ParseException.new( "Missing end tag for #{expanded_name}", 
+						source, self )
 				end
 			else
 				listener.tag_end name
@@ -384,33 +407,61 @@ module REXML
 
 
 		def parse source
-			# Get the next tag
-			md = source.match(/^<([\w-]*:?[\w-]+)(.*?>)/m, true)
-			self.name = md[1]
-			rest = md[2]
-			attrs = rest.scan(Attribute::PATTERN)
-			attrs.each do |set|
-				@attributes.add( Attribute.new( *set ) )
-			end
+			debug = false
+			begin
+				# Get the next tag
+				md = source.match(/^<((?:[:!\w_][-\.\w_]*:)?[-!\*\.\w_]*)\s*((((["']).*?\5)|[^\/'">]*)*?)(\/)?(>)/um, true)
+				raise ParseException.new("malformed XML: missing tag start", source,
+					self) unless md
+				puts "TAG: Matched \"#{md[0]}\", buffer is \"#{source.buffer}\"" if debug
+				puts "MATCH DATA = [#{md.to_a.join ', '}]" if debug
+				self.name = md[1]
+				attributes = md[2]
+				slash = md[6]
+				close = md[7]
+				if attributes
+					attrs = attributes.scan( Attribute::PATTERN )
+					raise ParseException.new( "error parsing attributes: [#{attrs.join ', '}], excess = \"#$'\"", source, self ) if $' and $'.strip.size > 0
+					#puts "ATTRIBUTES: Matched [#{attrs.join ', '}], buffer is #{source.buffer}" if debug
+					attrs.each do |set|
+						@attributes.add( Attribute.new( set[0], set[2] ) )
+					end
+				end
 
-			unless rest =~ /\/>$/		# empty
-				parse_children source
-				parse_error source, "Missing end tag for #{expanded_name}" unless source.match(/^\s*<\/#@expanded_name\s*>/m, true)
+				puts "CLOSE TAG: Matched #{close}, slash is #{slash}" if debug
+				if close and not slash
+					parse_children source
+					raise ParseException.new( "Missing end tag for #{expanded_name}",
+						source, self) unless source.match(/^\s*<\/#@expanded_name\s*>/um, 
+						true)
+				else
+					raise ParseException.new( "Malformed tag for #{expanded_name}",
+						source, self) unless close
+				end
+				puts "I AM NOW #{self}" if debug
+			rescue ParseException
+				$!.source = source
+				$!.element = self
+				raise
+			rescue Exception
+				old_ex = $!
+				raise ParseException.new("unidentified error", source, self, old_ex)
 			end
 		end
 
 		# A private helper method
 		def parse_children source
 			while true
-				md = source.match(/\A(\s*[^>]*>)/m)
-				parse_error source unless md
+				md = source.match(/\A(\s*[^>]*>)/um)
+				#puts "parse_children, match is #{md[0]}, rest is #{source.buffer}"
+				raise ParseException.new( "missing tag start", source, self ) unless md
 				line = md[1]
 				case line
 				when nil
 					return
-				when /\A<\w/m
-					add(Element.new(source))
-				when /\A<\//m
+				when /\A<[\w_:]/um
+					Element.new(source, self, @context)
+				when /\A<\//um
 					return
 				when CData::START_RE
 					add(CData.new(source))
@@ -419,7 +470,7 @@ module REXML
 				when Instruction::START_RE
 					add(Instruction.new(source))
 				else
-					add(Text.new(source))
+					add(Text.new(source, whitespace(), raw()))
 				end
 			end
 		end
@@ -440,16 +491,6 @@ module REXML
 				indent( writer, indent )
 			end
 		end
-
-		def parse_error source, message=nil
-			if message
-				error = "#{message}\n"
-			else
-				error = ""
-			end
-			error << "parse error at about:\n#{source.buffer[0..100]}\n"
-			raise error
-		end
 	end
 
 	########################################################################
@@ -517,14 +558,14 @@ module REXML
 		# the element is not a child
 		# @param element a child of this Element
 		def index element
-		  rv = 0
-		  found = @element.find do |child| 
-			 child.kind_of? Element and
-			 (rv += 1) and
-			 child == element
-		  end
-		  return rv if found == element
-		  return -1
+			rv = 0
+			found = @element.find do |child| 
+				child.kind_of? Element and
+				(rv += 1) and
+				child == element
+			end
+			return rv if found == element
+			return -1
 		end
 
 		# Deletes a child Element
@@ -560,12 +601,13 @@ module REXML
 		# @return the added Element
 		def add element=nil
 			if element.nil?
-				element = Element.new
+				Element.new "", self, @element.context
 			elsif not element.kind_of?(Element)
-				element = Element.new element
+				Element.new element, self, @element.context
+			else
+				@element << element
+				element.context = @element.context
 			end
-			@element << element
-			return element
 		end
 
 		alias :<< :add
@@ -586,11 +628,17 @@ module REXML
 			@element.each {|child| count+=1 if child.kind_of? Element }
 			count
 		end
+		
+		def to_a( xpath=nil )
+			rv = []
+			each( xpath ) { |el| rv << el }
+			rv
+		end
 
 		private
 		# A private helper method
 		def literalize name
-			literal = !(name =~ /^'.*'$/).nil?
+			literal = !(name =~ /^'.*'$/u).nil?
 			name = name[1..-2] if literal
 			[literal, name]
 		end
@@ -619,15 +667,26 @@ module REXML
 			return nil
 		end
 
+		def each_attribute
+			each_value do |val|
+				if val.kind_of? Attribute
+					yield val
+				else
+					val.each_value { |atr| yield atr }
+				end
+			end
+		end
+
 		# Fetches an attribute
 		# @param name the XPath by which to search for the attribute
 		# @return The first matching Attribute, or nil if there was none
 		def get_attribute( name )
 			attr = fetch( name, nil )
 			if attr.nil?
+				return nil if name.nil?
 				# Look for prefix
 				if name.include? ':'
-					name =~ /(\w+):(\w+)/
+					name =~ /(\w+):(\w+)/u
 					prefix, name = $1, $2
 					attr = fetch( name, nil )
 					# check prefix
@@ -665,7 +724,7 @@ module REXML
 			elsif old_attr.prefix != value.prefix
 				hash = { old_attr.prefix	=> old_attr,
 							value.prefix		=> value }
-				store value.name, hash
+							store value.name, hash
 			else
 				store value.name, value
 			end
@@ -680,18 +739,6 @@ module REXML
 			ns
 		end
 
-		def each_attribute
-			each_value do |attribute|
-				if attribute.kind_of? Hash
-					attribute.each_value do |attr|
-						yield attr
-					end
-				else
-					yield attribute
-				end
-			end
-		end
-
 		# Removes an attribute
 		# @return the removed attribute
 		def delete( attribute )
@@ -725,7 +772,7 @@ module REXML
 		private
 		# Private helper class.
 		def literalize name
-			literal = !(name =~ /^'.*'$/).nil?
+			literal = !(name =~ /^'.*'$/u).nil?
 			name = name[1..-2] if literal
 			[literal, name]
 		end
