@@ -4,8 +4,10 @@ module REXML
 		include Comparable
 
 		# string is String content, raw is a boolean
-		attr_accessor :string, :raw
-		PATTERN = /\A([^<]*)/m
+		attr_reader :string
+		attr_accessor :raw
+		PATTERN = /\A([^<]*)/um
+		ILLEGAL = /<|&(?!#?[\w-]+;)/u
 
 		# Constructor
 		# @param arg if a String, the content is set to the String.  If a Text,
@@ -15,21 +17,35 @@ module REXML
 		# @param whitespace (boolean, optional) if true, whitespace is respected
 		# @param parent (Parent, optional) if given, the parent of this object
 		# will be set to the argument
-		def initialize arg, whitespace=false, parent=nil, pattern=PATTERN
-			@raw = false
+		def initialize(arg, respect_whitespace=false, parent=nil, pattern=PATTERN, raw=false)
+			@raw = raw
+			if parent
+				if parent == true
+					@raw = true
+				elsif parent.kind_of? Parent
+					super( parent )
+					@raw = parent.raw
+				end
+			end
 			if arg.kind_of? Source
-				super( parent )
-				@string = arg.match(pattern, true)[1]
-				@string.squeeze!(" \n\t") unless whitespace
+				md = arg.match(pattern, true)
+				raise "no text to add" if md[0].length == 0
+				@string = md[1]
+				@string.squeeze!(" \n\t") unless respect_whitespace
+				@string = read_with_substitution(@string, ILLEGAL) unless @raw
 			elsif arg.kind_of? String
 				super()
-				@string = arg
+				@string = arg.clone
+				@string.squeeze!(" \n\t") unless respect_whitespace
+				@string = read_with_substitution(@string) unless @raw
 			elsif arg.kind_of? Text
 				super()
 				@string = arg.string
 				@raw = arg.raw
 			elsif arg.kind_of? Parent
 				super( arg )
+			elsif [true,false].include? arg
+				@raw = arg
 			end
 		end
 
@@ -50,8 +66,8 @@ module REXML
 
 		def write( writer, indent=0 ) 
 			indent( writer, indent )
-			if raw
-				writer.printf @string
+			if @raw
+				writer << @string
 			else
 				write_with_substitution( writer, @string )
 			end
